-- heightmap
-- a diamond-square algorithm for Lua
-- modified from https://github.com/mlepage/heightmap (retrieved January 18, 2018)


--[[
The MIT License (MIT)

Copyright (c) 2011 Marc Lepage

Permission is hereby granted, free of charge, to any person obtaining a copy
of this software and associated documentation files (the "Software"), to deal
in the Software without restriction, including without limitation the rights
to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
copies of the Software, and to permit persons to whom the Software is
furnished to do so, subject to the following conditions:

The above copyright notice and this permission notice shall be included in
all copies or substantial portions of the Software.

THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN
THE SOFTWARE.
]]--


local M = {}

local abs, noise, pi, random, sin, sqrt = math.abs, love.math.noise, math.pi, love.math.random, math.sin, math.sqrt

-- Noise function
local function noisefn(h, d)
	return (h + (random() - 0.5) * d)
end

-- Square step
-- Sets map[x][y] from square of radius d using height function f
local function square(map, x, y, d)
	local sum, num = 0, 0
	if x-d > 0 then
		if y-d > 0      then sum, num = sum + map[x-d][y-d], num + 1 end
		if y+d <= map.h then sum, num = sum + map[x-d][y+d], num + 1 end
	end
	if x+d <= map.w then
		if y-d > 0      then sum, num = sum + map[x+d][y-d], num + 1 end
		if y+d <= map.h then sum, num = sum + map[x+d][y+d], num + 1 end
	end
	map[x][y] = noisefn(sum/num, d)
end

-- Diamond step
-- Sets map[x][y] from diamond of radius d using height function f
local function diamond(map, x, y, d)
	local sum, num = 0, 0
	if x-d > 0      then sum, num = sum + map[x-d][y], num + 1 end
	if x+d <= map.w then sum, num = sum + map[x+d][y], num + 1 end
	if y-d > 0      then sum, num = sum + map[x][y-d], num + 1 end
	if y+d <= map.h then sum, num = sum + map[x][y+d], num + 1 end
	map[x][y] = noisefn(sum/num, d)
end

-- Diamond square algorithm generates cloud/plasma fractal heightmap
-- http://en.wikipedia.org/wiki/Diamond-square_algorithm
-- Size must be power of two
-- Height function f must look like f(map, x, y, d, h) and return h'
local function diamondSquare(size)
	-- initialize map with zeros
	local map = { w = size, h = size }
	for x = 0, size do
		map[x] = {}
		for y = 0, size do
			map[x][y] = 0
		end
	end
	-- seed four corners
	map[0][0] = noisefn(0, size)
	map[size][0] = noisefn(0, size)
	map[0][size] = noisefn(0, size)
	map[size][size] = noisefn(0, size)
	-- perform square and diamond steps
	local step = size
	repeat
		step = step / 2
		for x = step, map.w - 1, step * 2 do
			for y = step, map.h - 1, step * 2 do
				square(map, x, y, step, ID)
			end
		end
		for x = step, map.w - 1, step * 2 do
			for y = 0, map.h, step * 2 do
				diamond(map, x, y, step, ID)
			end
		end
		for x = 0, map.w, step * 2 do
			for y = step, map.h - 1, step * 2 do
				diamond(map, x, y, step, ID)
			end
		end
	until step <= 1
	return map
end

function M.create(width, height)
	-- find containing square
	local square = 1
	repeat
		square = square * 2
	until square >= width and square >= height
	-- generate diamond-square heightmap
	local map = diamondSquare(square, ID)
	-- clip heightmap to requested size
	for x = 0, map.w do
		for y = height+1, map.h do
			map[x][y] = nil
		end
	end
	for x = width+1, map.w do
		map[x] = nil
	end
	-- normalize heightmap
	map.w = width
	map.h = height
	local min, max = map[1][1], map[1][1]
	for x = 1, map.w do
		for y = 1, map.h do
			min = map[x][y] < min and map[x][y] or min
			max = map[x][y] > max and map[x][y] or max
		end
	end
	for x = 1, map.w do
		for y = 1, map.h do
			map[x][y] = (max-map[x][y]) / (max-min)
		end
	end
	-- return heightmap
	return map
end

return M
