
-- CotW map functions

local M = {}

-- UTILITY FUNCTIONS --

function smooth(map, range)
	local smoothed = {}
	for x = 1, #map do
		smoothed[x] = {}
		for y = 1, #map[x] do
			local sum, count = 0, 0
			for i = x-range, x+range do
				for j = y-range, y+range do
					local val = 2^((range*2) - math.abs(x-i) - math.abs(y-j))
					sum = sum + (map[i] and map[i][j] or 0) * val -- returns 0 for an out-of-bounds read without crashing :)
					count = count + val
				end
			end
			smoothed[x][y] = sum/count
		end
	end
	return smoothed
end

-- MAP CREATION FUNCTIONS --

function cellularAutomata(mapWidth, mapLength)

	-- seed map
	local solidChance, map = 0.55, {}
	for x = 1, mapWidth do
		map[x] = {}
		for y = 1, mapLength do
			map[x][y] = love.math.random() < solidChance and 1 or 0
		end
	end

	-- cellular automata process (using 4 generations of the 4-5 rule)
	for gen = 1, 4 do
		for x = 1, mapWidth do
			for y = 1, mapLength do
				-- count current neighbors
				local neighborCount = 0
				for i = math.max(x-1,1), math.min(x+1,mapWidth) do
					for j = math.max(y-1,1), math.min(y+1,mapLength) do
						neighborCount = neighborCount + map[i][j]
					end
				end
				map[x][y] = neighborCount > 4 and 1 or 0
			end
		end
	end

	return map
end

function M.createNewWorld(mapWidth, mapLength)


	-- MAP TERRAIN

	-- Create diamond-square heightmap
	local mapHeight = diamondsquare.create(mapWidth, mapLength)

	-- Apply cellular automata terrain mask
	local mapMask = cellularAutomata(mapWidth, mapLength)
	for x = 1, mapWidth do
		for y = 1, mapLength do
			mapHeight[x][y] = mapHeight[x][y] * mapMask[x][y]
		end
	end
	mapMask = nil

	-- Smooth world
	mapHeight = smooth(mapHeight, 4)
	mapHeight = smooth(mapHeight, 2)
	mapHeight = smooth(mapHeight, 1)

	-- Create rainfall map
	local mapRain = diamondsquare.create(mapWidth, mapLength, 1)
	mapRain = smooth(mapRain, 3)

	-- Create rivers
	local HEIGHT = {Water=0.2, Ground=0.65, Spring=0.63, NewRiver=1.0}
	local RAIN = {Desrt=0.2, Grass=0.6, Trees=0.85, Spring=0.93}
	for x = 1, mapWidth do
		for y = 1, mapLength do
			-- rivers start at high-altitude rainy places
			if mapHeight[x][y] >= HEIGHT.Spring and mapRain[x][y] > RAIN.Spring then
				local river = {x=x, y=y, z, dx, dy}
				river[1] = {x=x, y=y}
				repeat
					-- run the river through lowest neighbor until water or map edge
					mapHeight[river.x][river.y] = HEIGHT.NewRiver
					river.z = HEIGHT.NewRiver -- this is different than .Water so the river doesn't end when it touches itself
					if river.x > 1 and mapHeight[river.x-1][river.y] < river.z then
						river.dx = -1; river.dy = 0; river.z = mapHeight[river.x-1][river.y] end
					if river.y > 1 and mapHeight[river.x][river.y-1] < river.z then
						river.dx = 0; river.dy = -1; river.z = mapHeight[river.x][river.y-1] end
					if river.x < mapWidth and mapHeight[river.x+1][river.y] < river.z then
						river.dx = 1; river.dy = 0; river.z = mapHeight[river.x+1][river.y] end
					if river.y < mapLength and mapHeight[river.x][river.y+1] < river.z then
						river.dx = 0; river.dy = 1; river.z = mapHeight[river.x][river.y+1] end
					if river.dx == nil then -- if the river gets trapped, make deep water
						mapHeight[river.x][river.y] = 0
					else
						river.x = river.x + river.dx
						river.y = river.y + river.dy
						river[#river+1] = {x = river.x, y = river.y}
					end
				until mapHeight[river.x][river.y] <= HEIGHT.Water or river.x == 1 or river.x == mapWidth or river.y == 1 or river.y == mapLength
				if mapHeight[river.x][river.y] <= HEIGHT.Water then river[#river] = nil end -- don't overwrite existing water
				for i = 1, #river do -- apply the defined river
					mapHeight[river[i].x][river[i].y] = HEIGHT.Water
				end
			end
		end
	end

	-- Create Terrain
	local terrainCost = {6.5,  2.6,  1.4,  1.0,  1.8,  2.2,  6.5}
	local map = {w=mapWidth, h=mapLength, terDefault=terOcean, roadCost=0.1, avgCost=0.5}
	for x = 1, mapWidth do
		map[x] = {}
		for y = 1, mapLength do
			map[x][y] = {}
			local height, rain = mapHeight[x][y], mapRain[x][y]
			if     height  < HEIGHT.Water then terType = terOcean
			elseif height == HEIGHT.Water then terType = terRiver
			elseif height  < HEIGHT.Ground then
				if     rain < RAIN.Desrt then terType = terDesrt
				elseif rain < RAIN.Grass then terType = terGrass
				elseif rain < RAIN.Trees then terType = terTrees
				else   terType = terSwamp
				end
			else
				terType = terRocks
			end
			map[x][y].feature = 0
			map[x][y].terrain = terType
			map[x][y].pathCost = terrainCost[terType]
		end
	end
	mapRain = nil


	-- WORLD MAP LOCATIONS

	-- Initialize
	local terrainVal = {12, 6, 1, 3, 6, 2, 9}

	-- Initial location nodes
	local nodeByID = {count = math.floor((mapWidth*mapLength)^0.25)}
	for n = 1, nodeByID.count do
		nodeByID[n] = {x=math.random(1, mapWidth), y=math.random(1, mapLength), size=0, val=0}
	end

	-- Voroni regions
	local nodeByCell = {}
	for x = 1, mapWidth do
		nodeByCell[x] = {}
		for y = 1, mapLength do
			if map[x][y].terrain > terRiver and map[x][y].terrain < terRocks then
				local ID, dist = 0, mapWidth+mapLength
				for n = 1, nodeByID.count do
					if math.abs(nodeByID[n].x - x) + math.abs(nodeByID[n].y - y) < dist then
						ID = n
						dist = math.abs(nodeByID[n].x - x) + math.abs(nodeByID[n].y - y)
					end
				end
				nodeByCell[x][y] = ID
			end
		end
	end

	-- Lloyd relaxation
	for n = 1, nodeByID.count do
		local sum = {x=0, y=0, count=0}
		for x = 1, mapWidth do
			for y = 1, mapLength do
				if nodeByCell[x][y] == n then
					sum.count = sum.count + 1
					sum.x = sum.x + x
					sum.y = sum.y + y
				end
			end
		end
		nodeByID[n].x = math.floor(sum.x/sum.count)
		nodeByID[n].y = math.floor(sum.y/sum.count)
	end

	-- Update Voroni Regions
	for x = 1, mapWidth do
		for y = 1, mapLength do
			if map[x][y].terrain > terRiver and map[x][y].terrain < terRocks then
				local ID, dist = 0, mapWidth+mapLength
				for n = 1, nodeByID.count do
					if math.abs(nodeByID[n].x - x) + math.abs(nodeByID[n].y - y) < dist then
						ID = n
						dist = math.abs(nodeByID[n].x - x) + math.abs(nodeByID[n].y - y)
					end
				end
				nodeByCell[x][y] = ID
				nodeByID[ID].size = nodeByID[ID].size + 1
			end
		end
	end

	-- Normalize region sizes
	local max, min = 0, mapWidth*mapLength
	for n = 1, nodeByID.count do
		max = nodeByID[n].size > max and nodeByID[n].size or max
		min = nodeByID[n].size < min and nodeByID[n].size or min
	end
	for n = 1, nodeByID.count do
		nodeByID[n].size = (nodeByID[n].size - min) / (max-min)
	end

	-- Place Location at highest-value location in each Region
	nodeByID.avgVal = 0
	for n = 1, nodeByID.count do
		local locationRange = 3
		local Location = {x=0,y=0,val=0}
		for x = 1, mapWidth do
			for y = 1, mapLength do
				if nodeByCell[x][y] == n then -- there's probably a better way than this...
					local value = 0
					for i = math.max(x-locationRange,1), math.min(x+locationRange,mapWidth) do
						for j = math.max(y-locationRange,1), math.min(y+locationRange,mapLength) do
							value = value + terrainVal[map[i][j].terrain]
							if nodeByCell[i][j] == n then value = value + terrainVal[map[i][j].terrain] end -- double points for cells in your OWN region
						end
					end
					if value > Location.val then
						Location.x = x
						Location.y = y
						Location.val = value
					end
				end
			end
		end
		nodeByID[n].x = Location.x
		nodeByID[n].y = Location.y
		nodeByID[n].val = Location.val
		nodeByID.avgVal = nodeByID.avgVal + Location.val/nodeByID.count
		map[Location.x][Location.y].pathCost = map.roadCost
	end


	-- LOCATION CONNECTING (ROAD BUILDING ETC)

	-- https://www.raywenderlich.com/4946/introduction-to-a-pathfinding
	-- ***
	-- for n1 = 1, LocationCount-1 do
	-- 		for n2 = n1 + 1, LocationCount do
	--			PATHFIND(n1, n2)
	--			Remember, when you GET NEIGHBORS, a Tunnl or Ferry MUST also return its opposite end!!  Cost is (.road * Dist)
	--			Non-road features (guarantee accessibility)
	--				River:
	--					build a Bridge (Feature? Terrain?)
	--				Ocean:
	-- 					1 or 2 tiles in length = bridge
	-- 					else Ferry endpoints, nothing in middle
	--				Rocks:
	-- 					1 or 2 tiles in length = whatever terrain is under endpoints
	-- 					else Tunnl endpoints, nothing in middle
	--			Build roads:
	-- 				City points  50   GUESSES ONLY - test & tweak!
	-- 				Town points  20
	-- 				Hmlt points   9
	-- 				EVIL points   0
	--				RoadPoints = n1.Points * n2.Points
	--				If RoadPoints >= RoadCost then BuildRoad!
	--				New road = SMALL road
	--				pathing over existing roadd = upgrade to LARGE road
	--				don't overwrite other map features, lol!
	--		end
	-- end


	-- Calculate Manhattan distances between Locations
	-- HACK *** Use this or skip it?  It's an effort to build cheaper roads first, make pathfinding more efficient, etc.
	local nodeDist = {}
	for n1 = 1, nodeByID.count - 1 do
		for n2 = n1 + 1, nodeByID.count do
			nodeDist[#nodeDist + 1] = {}
			nodeDist[#nodeDist].x1 = nodeByID[n1].x
			nodeDist[#nodeDist].y1 = nodeByID[n1].y
			nodeDist[#nodeDist].x2 = nodeByID[n2].x
			nodeDist[#nodeDist].y2 = nodeByID[n2].y
			nodeDist[#nodeDist].dist = math.abs(nodeByID[n1].x - nodeByID[n2].x) + math.abs(nodeByID[n1].y - nodeByID[n2].y)
		end
	end
	-- Sort by Ascending distance
	for i = 1, #nodeDist - 1 do
		for j = i+1, #nodeDist do
			if nodeDist[i].dist > nodeDist[j].dist then
				local tmp = nodeDist[j]
				nodeDist[j] = nodeDist[i]
				nodeDist[i] = tmp
			end
		end
	end

	-- Calculate paths between Locations
	local paths = {}
	for i = 1, #nodeDist do
		paths[i] = ai.pathfind(map, nodeDist[i].x1, nodeDist[i].y1, nodeDist[i].x2, nodeDist[i].y2)
		if paths then
			local roadType = nil
			for j = 1, #paths[i] do
				local x = paths[i][j].x
				local y = paths[i][j].y
				map[x][y].pathCost = map.roadCost
				map[x][y].feature = map[x][y].feature + 1
			end
		end
	end

	-- Set road tiles, etc
	for x = 1, mapWidth do
		for y = 1, mapLength do
			-- HACK *** Account for point differences between hamlets, towns, cities
			-- HACK *** Apply tunnels & ferries as needed
			if map[x][y].feature > 0 then
				local roadShape = 19
				if map[x][y].feature > 1 then roadShape = roadShape + 18 end -- HACK *** Automate this ?
				if x > 1 and map[x-1][y].feature > 0 then roadShape = roadShape + 1 end
				if y > 1 and map[x][y-1].feature > 0 then roadShape = roadShape + 2 end
				if x < mapWidth  and map[x+1][y].feature > 0 then roadShape = roadShape + 4 end
				if y < mapLength and map[x][y+1].feature > 0 then roadShape = roadShape + 8 end
				map[x][y].feature = roadShape
			end
		end
	end

	-- Clear unused flags
	for x = 1, mapWidth do
		for y = 1, mapLength do
			if map[x][y].feature == 0 then map[x][y].feature = nil end
		end
	end

	-- Set Location graphics
	for n = 1, nodeByID.count do
		local x, y = nodeByID[n].x, nodeByID[n].y
		if nodeByID[n].val > nodeByID.avgVal then
			if     nodeByID[n].size < 0.33 then map[x][y].feature = ftrHmlt
			elseif nodeByID[n].size < 0.67 then map[x][y].feature = ftrTown
			else map[x][y].feature = ftrCity
			end
			player.x = x -- HACK *** Should be done elsewhere / later ?
			player.y = y
		else
			if     nodeByID[n].size < 0.33 then map[x][y].feature = ftrCave
			elseif nodeByID[n].size < 0.67 then map[x][y].feature = ftrRuin
			else map[x][y].feature = ftrDngn
			end
		end
	end

	-- LOCATION GENERATION
	-- ***
	-- City gen
	-- ...
	-- https://procworld.blogspot.com/2011/07/city-lots.html?m=1
	-- ...


	-- Starting location
	-- ***
	-- Start the player in a Hamlet, if one exists
	-- If not, a Town
	-- Else, a City
	return {map} -- HACK *** Return all maps here, with world map being [1] etc
end

return M
